#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import ast
import inspect
from pprint import pformat

from .isinstance import *

__all__ = ["dbg"]


class FmtKeyError(Exception):
    def __init__(self, keys):
        msg = 'fmt error; key not found in keys: ' + ' '.join(keys)
        super(FmtKeyError, self).__init__(msg)

def dbg(*args, **kwargs):
    logger = kwargs.pop("logger", None)
    frame = inspect.currentframe().f_back
    return _dbg(args, kwargs, frame, logger=logger)


def _create_format(name):
    try:
        return str(ast.literal_eval(name))
    except Exception:
        return name + "({" + name + '_type})="{' + name + '}"'


_dbg_regex = re.compile(r"p?dbg\s*\((.+?)\)$")


def _dbg(args, kwargs, frame, logger=None):
    caller = inspect.getframeinfo(frame)
    string = f"DBG:{os.path.relpath(caller.filename)}:{caller.lineno} "
    instance = frame.f_locals.get("self", None)
    if instance:
        string += instance.__class__.__name__ + "."
    string += frame.f_code.co_name + "(): "
    if args or kwargs:
        context = inspect.getframeinfo(frame).code_context
        callsite = "".join([line.strip() for line in context])
        match = _dbg_regex.search(callsite)
        if match:
            params = [
                param.strip() for param in match.group(1).split(",") if "=" not in param
            ]
        names = params[: len(args)] + list(kwargs.keys())
        string += " ".join([_create_format(name) for name in names])
        arg_types = {name + "_type": type(arg) for name, arg in zip(names, args)}
        kwarg_types = {key + "_type": type(value) for key, value in kwargs.items()}
        kwargs.update(arg_types)
        kwargs.update(kwarg_types)
    else:
        string += "locals:\n{locals}"
    result = _fmt(string, args, kwargs, frame, do_print=logger is None)
    if logger:
        logger.debug(result)
    return result


def _fmt_dict(obj):
    if isdict(obj) or islist(obj):
        return pformat(obj)
    return str(obj)


def _fmt(string, args, kwargs, frame, do_print=False):
    try:
        gl = {"locals": frame.f_locals, "globals": frame.f_globals}
        gl.update(frame.f_globals)
        gl.update(frame.f_locals)
        gl.update(kwargs)
        if frame.f_code.co_name == "<listcomp>":
            gl.update(frame.f_back.f_locals)
        result = string.format(*args, **{k: _fmt_dict(v) for k, v in gl.items()})
    except KeyError as ke:
        raise FmtKeyError(gl)
    if do_print:
        print(result)
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
collection of isinstance funcs
"""

from collections import Iterable
from types import ModuleType, FunctionType, GeneratorType


def isstr(obj):
    """
    isstr
    """
    return isinstance(obj, str)


def isint(obj):
    """
    isint
    """
    return isinstance(obj, int)


def isfloat(obj):
    """
    isfloat
    """
    return isinstance(obj, float)


def ismodule(obj):
    """
    ismodule
    """
    return isinstance(obj, ModuleType)


def isfunction(obj):
    """
    isfunction
    """
    return isinstance(obj, FunctionType)


def isgenerator(obj):
    """
    isgenerator
    """
    return isinstance(obj, GeneratorType)


def isscalar(obj):
    """
    isscalar
    """
    return obj is None or isstr(obj) or isint(obj) or isfloat(obj)


def islist(obj):
    """
    islist
    """
    return isinstance(obj, list)


def istuple(obj):
    """
    istuple
    """
    return isinstance(obj, tuple)


def isiterable(obj):
    """
    isiterable
    """
    return isinstance(obj, Iterable)


def isdict(obj):
    """
    isdict
    """
    return isinstance(obj, dict)


def isa(obj, *types, is_all=False):  # noqa: E999
    if is_all:
        return all([isinstance(obj, t) for t in types])
    return any([isinstance(obj, t) for t in types])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


class Singleton(object):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._instance



"""
git utilities
"""

from .shell import call, cd


def reporoot(path="."):
    with cd(path):
        return call("git rev-parse --show-toplevel")[1].strip()


def describe(path="."):
    with cd(path):
        return call("git describe --abbrev=7")[1].strip()


def subs2shas(path="."):
    lines = call("cd %s && git submodule" % path)[1].strip().split("\n")
    return dict([(item[1], item[0]) for item in [line.split() for line in lines]])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
shell utilities
"""

import os
import re
import sys
import glob
import fnmatch
from subprocess import Popen, PIPE, CalledProcessError, check_output
from contextlib import contextmanager

# pylint: disable=invalid-name


def expandpath(path):
    return os.path.realpath(os.path.expanduser(path))


def inversepath(path):
    return "/".join([".." for _ in path.split("/")])


def which(string, **kwargs):
    output = check_output("which " + string, shell=True, **kwargs)
    return output.decode("utf-8").strip()


@contextmanager
def cd(*args, **kwargs):
    mkdir = kwargs.pop("mkdir", True)
    verbose = kwargs.pop("verbose", False)
    path = os.path.sep.join(args)
    path = os.path.normpath(path)
    path = os.path.expanduser(path)
    prev = os.getcwd()
    if path != prev:
        if mkdir:
            call("mkdir -p %(path)s" % locals(), verbose=verbose)
        os.chdir(path)
        curr = os.getcwd()
        sys.path.append(curr)
        if verbose:
            print("cd %s" % curr)
    try:
        yield
    finally:
        if path != prev:
            sys.path.remove(curr)
            os.chdir(prev)
            if verbose:
                print("cd %s" % prev)


def call(
    cmd, stdout=PIPE, stderr=PIPE, shell=True, nerf=False, throw=True, verbose=False
):
    if verbose or nerf:
        print(cmd)
    if nerf:
        return (None, "nerfed", "nerfed")
    process = Popen(cmd, stdout=stdout, stderr=stderr, shell=shell)
    _stdout, _stderr = [stream.decode("utf-8") for stream in process.communicate()]
    exitcode = process.poll()
    if verbose:
        if _stdout:
            print(_stdout)
        if _stderr:
            print(_stderr)
    if throw and exitcode:
        raise CalledProcessError(
            exitcode, "cmd=%(cmd)s; stdout=%(_stdout)s; stderr=%(_stderr)s" % locals()
        )
    return exitcode, _stdout, _stderr


def rglob(pattern):
    matches = []
    # support for shell-like {x,y} syntax
    regex = re.compile("(.*){(.*)}(.*)")
    match = regex.search(pattern)
    if match:
        prefix, alternates, suffix = match.groups()
        for alternate in alternates.split(","):
            matches += rglob(prefix + alternate.strip() + suffix)
        return matches
    # support for recursive glob
    for r, _, fs in os.walk(os.path.dirname(pattern)):
        for f in fnmatch.filter(fs, os.path.basename(pattern)):
            matches.append(os.path.join(r, f))
    return matches


def globs(recursive=False, *paths):
    """
    returns a set of all paths glob-matched
    """
    return set(
        [
            item
            for item in [glob.glob(path, recursive=recursive) for path in paths]
            for item in item
        ]
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import imp


def import_modules(dirpath, endswith):
    files = [f for f in os.listdir(dirpath) if f.endswith(endswith)]
    return [imp.load_source(f.split(".")[0], dirpath + "/" + f) for f in files]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
from enum import Enum
from fnmatch import fnmatch
from collections import OrderedDict


class MatchType(Enum):
    EXACT = 0
    IGNORECASE = 1
    PREFIX = 2
    SUFFIX = 3
    CONTAINS = 4
    GLOB = 5
    REGEX = 6


class InvalidFuzzyTypeError(Exception):
    def __init__(self, obj):
        message = f"type(obj)={type(obj)} is not dict or list"
        super(InvalidFuzzyTypeError, self).__init__(message)


MATCH_FUNCS = {
    MatchType.EXACT: lambda item, pattern: pattern == item,
    MatchType.IGNORECASE: lambda item, pattern: pattern.lower() == item.lower(),
    MatchType.PREFIX: lambda item, pattern: fnmatch(item, pattern + "*"),
    MatchType.SUFFIX: lambda item, pattern: fnmatch(item, "*" + pattern),
    MatchType.CONTAINS: lambda item, pattern: pattern in item,
    MatchType.GLOB: lambda item, pattern: fnmatch(item, pattern),
    MatchType.REGEX: lambda item, pattern: re.search(item, pattern),
}

DEFAULT_MATCH_TYPES = [
    MatchType.EXACT,
    MatchType.IGNORECASE,
    MatchType.PREFIX,
    MatchType.CONTAINS,
]


def match_item(item, patterns, match_type, include):
    match_func = MATCH_FUNCS[match_type]
    if include:
        return any([match_func(item, pattern) for pattern in patterns])
    return all([not match_func(item, pattern) for pattern in patterns])


def match_items(items, patterns, match_types, include=True):
    for match_type in match_types:
        results = [
            item for item in items if match_item(item, patterns, match_type, include)
        ]
        if results:
            return results
    return []

class FuzzyTuple(tuple):
    def __new__(cls, *args, **kwargs):
        self._type = type(args[0])
        return super().__new__(cls, tuple(*args))

    def __init__(self, *args, **kwargs):
        self._match_types = kwargs.pop("match_types", DEFAULT_MATCH_TYPES)

    def include(self, *patterns, match_types=None):
        items = match_items(
            [item for item in self.__iter__()],
            patterns,
            match_types or self._match_types,
            include=True,
        )
        return FuzzyTuple(items)

    def exclude(self, *patterns, match_types=None):
        items = match_items(
            [item for item in self.__iter__()],
            patterns,
            match_types or self._match_types,
            include=False,
        )
        return FuzzyTuple(items)

    def defuzz(self):
        return tuple(item for item in self.__iter__())

    def __repr__(self):
        return repr(tuple(self))


class FuzzyList(list):
    def __init__(self, *args, **kwargs):
        self._type = type(args[0])
        self._match_types = kwargs.pop("match_types", DEFAULT_MATCH_TYPES)
        super().__init__(*args, **kwargs)

    def include(self, *patterns, match_types=None):
        items = match_items(
            [item for item in self.__iter__()],
            patterns,
            match_types or self._match_types,
            include=True,
        )
        return FuzzyList(items)

    def exclude(self, *patterns, match_types=None):
        items = match_items(
            [item for item in self.__iter__()],
            patterns,
            match_types or self._match_types,
            include=False,
        )
        return FuzzyList(items)

    def defuzz(self):
        return [item for item in self.__iter__()]


class FuzzyDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        self._type = type(args[0])
        super().__init__(*args, **kwargs)

    def include(self, *patterns, match_types=None):
        items = match_items(
            self.keys(), patterns, match_types or self._match_types, include=True
        )
        return FuzzyDict({item: self.get(item) for item in items})

    def exclude(self, *patterns, match_types=None):
        items = match_items(
            self.keys(), patterns, match_types or self._match_types, include=False
        )
        return FuzzyDict({item: self.get(item) for item in items})

    def defuzz(self):
        if self._type.__name__ == 'dict':
            return dict(self.items())
        elif self._type.__name__ == 'collections.OrderedDict':
            return OrderedDict(self.items())
        raise Exception(f"unknown type: {self._type}")


def fuzzy(obj):
    if isinstance(obj, tuple):
        return FuzzyTuple(obj)
    elif isinstance(obj, list):
        return FuzzyList(obj)
    elif isinstance(obj, dict):
        return FuzzyDict(obj)
    raise InvalidFuzzyTypeError(obj)


if __name__ == "__main__":
    f = fuzzy(["a", "b"])
    print(f"f={f.items}")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
url
"""
import socket

from urlpath import URL


class UrlDnsError(Exception):
    def __init__(self, url, errors=None):
        msg = f"url dns resoltuion with {url}"
        super(UrlDnsError, self).__init__(msg)


class UrlConnectError(Exception):
    def __init__(self, url, errors=None):
        msg = f"url connect error {url}"
        super(UrlConnectError, self).__init__(msg)


def validate(url, throw=False):
    if not isinstance(url, URL):
        url = URL(url)
    proto = url.components[0]
    hostname = url.hostname
    try:
        address = socket.gethostbyname(hostname)
    except Exception as ex:
        if throw:
            raise UrlDnsError(url, errors=[ex])
        return False
    try:
        family, socktype, proto, _, address = socket.getaddrinfo(address, proto)[0]
        s = socket.socket(family, socktype, proto)
        s.connect(address)
    except Exception as ex:
        if throw:
            raise UrlConnectError(url, errors=[ex])
        return False
    return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
config.py

"""

import os
import traceback
import configparser
from ruamel import yaml


class ConfigLoadError(Exception):
    def __init__(self, filename, errors):
        msg = f"config load error: filename={filename} errors={errors}"
        super(ConfigLoadError, self).__init__(msg)


def load_yaml_or_json(filename):
    with open(filename) as f:
        return yaml.safe_load(f)


def load_ini_or_cfg(filename):
    parser = configparser.ConfigParser()
    parser.read(filename)
    sections = {}
    for section in parser.sections():
        options = {}
        for option in parser.options(section):
            value = parser.get(section, option)
            options[option] = value
            sections[section] = options
    return sections


def load_config(*filenames, must_exist=False):
    cfg = {}
    errors = []
    for filename in [
        os.path.expanduser(filename) for filename in filenames if filename
    ]:
        if filename and os.path.isfile(filename):
            try:
                cfg.update(load_ini_or_cfg(filename))
            except Exception as ex1:
                errors += traceback.format_exc()
                try:
                    cfg.update(load_yaml_or_json(filename))
                except Exception as ex2:
                    errors += traceback.format_exc()
                    raise ConfigLoadError(filename, errors)
        elif must_exist:
            raise FileNotFoundError(filename)
    return cfg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
human friendly functions
"""

import re
import datetime

WEEKS = "(?P<weeks>[0-9]+w)?"
DAYS = "(?P<days>[0-9]+d)?"
HOURS = "(?P<hours>[0-9]+h)?"
MINUTES = "(?P<minutes>[0-9]+m)?"
SECONDS = "(?P<seconds>[0-9]+s)?"
PARSE = re.compile(WEEKS + DAYS + HOURS + MINUTES + SECONDS)
VALIDATE = re.compile("^([0-9]+(w|d|h|m|s))+$")


class FriendlyTimedeltaParseError(Exception):
    def __init__(self, string):
        msg = "friendly.timedelta parse error for string: " + string
        super(FriendlyTimedeltaParseError, self).__init__(msg)


def timedelta(string):
    if VALIDATE.match(string):
        match = PARSE.match(string)
        if not match:
            raise FriendlyTimedeltaParseError(string)
        kwargs = {k: int(v[:-1]) for k, v in match.groupdict().items() if v is not None}
        return datetime.timedelta(**kwargs)
    raise FriendlyTimedeltaParseError(string)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .git import describe
from .shell import cd


def get_version(path="."):
    try:
        try:
            with cd(path):
                value = open("VERSION").read().strip()
        except:  # noqa: E722
            value = describe()
    except:  # noqa: E722
        value = "UNKNOWN"
    version, *suffix = value.split("-")
    return version.replace("v", "") + (".dev{0}+{1}".format(*suffix) if suffix else "")


class Version(str):
    __instance = None

    def __new__(cls):
        if Version.__instance is None:
            value = get_version()
            Version.__instance = super(Version, cls).__new__(cls, value)
        return Version.__instance


version = Version()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
output
"""

from ruamel import yaml


def str_presenter(dumper, data):
    str_tag = "tag:yaml.org,2002:str"
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar(str_tag, data, style="|")
    return dumper.represent_scalar(str_tag, data)


yaml.add_representer(str, str_presenter)


def list_presenter(dumper, data):
    list_tag = "tag:yaml.org,2002:seq"
    if len(data) > 1:
        if all([isinstance(item, str) for item in data]):
            return dumper.represent_sequence(list_tag, data, flow_style=False)
    return dumper.represent_sequence(list_tag, data)


yaml.add_representer(list, list_presenter)


def yaml_format(obj):
    class MyDumper(yaml.Dumper):
        def represent_mapping(self, tag, mapping, flow_style=False):
            return yaml.Dumper.represent_mapping(self, tag, mapping, flow_style)

    return yaml.dump(obj, Dumper=MyDumper).strip()


def yaml_print(obj):
    return print(yaml_format(obj))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import hashlib
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.backends.openssl.rsa import _RSAPrivateKey
from cryptography.hazmat.backends.openssl.x509 import _CertificateSigningRequest


class CreateModhashRequiresKeyOrCsrError(Exception):
    def __init__(self, obj):
        msg = f"_create_modhash requires key or csr but got obj={obj}"
        super(CreateModhashRequiresKeyOrCsrError, self).__init__(msg)


ENCODING = dict(DER=serialization.Encoding.DER, PEM=serialization.Encoding.PEM)

OIDS_MAP = dict(
    common_name=NameOID.COMMON_NAME,
    org_name=NameOID.ORGANIZATION_NAME,
    org_unit=NameOID.ORGANIZATIONAL_UNIT_NAME,
    org_country=NameOID.COUNTRY_NAME,
    org_city=NameOID.LOCALITY_NAME,
    org_state=NameOID.STATE_OR_PROVINCE_NAME,
    org_zip=NameOID.POSTAL_CODE,
    org_addr1=NameOID.STREET_ADDRESS,
    org_addr2=NameOID.STREET_ADDRESS,
    org_contact_job_title=NameOID.TITLE,
    org_contact_firstname=NameOID.GIVEN_NAME,
    org_contact_lastname=NameOID.SURNAME,
    org_contact_email=NameOID.EMAIL_ADDRESS,
)


def _create_key(common_name, public_exponent, key_size, **kwargs):
    key = rsa.generate_private_key(
        public_exponent=public_exponent, key_size=key_size, backend=default_backend()
    )
    return key


def _create_oids(common_name, oids):
    attrs = [x509.NameAttribute(NameOID.COMMON_NAME, common_name)]
    for key, value in oids.items():
        attr = OIDS_MAP.get(key, None)
        if attr:
            attrs += [x509.NameAttribute(attr, str(value))]
    return attrs


def _add_sans(subject, sans):
    subject.add_extension(
        x509.SubjectAlternativeName([x509.DNSName(san) for san in sans]), critical=False
    )


def _create_csr(common_name, key, oids=None, sans=None):
    builder = x509.CertificateSigningRequestBuilder()
    oids = _create_oids(common_name, oids if oids else {})
    subject = builder.subject_name(x509.Name(oids))
    if sans:
        _add_sans(subject, sans)
    csr = subject.sign(key, hashes.SHA256(), default_backend())
    return csr


def _create_modhash(obj):
    if isinstance(obj, _RSAPrivateKey):
        modulus_int = obj.private_numbers().public_numbers.n
    elif isinstance(obj, _CertificateSigningRequest):
        modulus_int = obj.public_key().public_numbers().n
    else:
        raise CreateModhashRequiresKeyOrCsrError(obj)
    modulus_hex = hex(modulus_int).rstrip("L").lstrip("0x").upper()
    modulus_bytes = f"Modulus={modulus_hex}\n".encode("utf-8")
    md5 = hashlib.md5()
    md5.update(modulus_bytes)
    return md5.hexdigest()


def get_serial(crt):
    crt = x509.load_pem_x509_certificate(str.encode(crt), default_backend())
    return crt.serial_number


def get_sha1(crt):
    crt = x509.load_pem_x509_certificate(str.encode(crt), default_backend())
    return crt.fingerprint(hashes.SHA1()).hex()


def get_sha2(crt):
    crt = x509.load_pem_x509_certificate(str.encode(crt), default_backend())
    return crt.fingerprint(hashes.SHA256()).hex()


def create_modhash_key_and_csr(
    common_name,
    public_exponent=None,
    key_size=None,
    key=None,
    csr=None,
    oids=None,
    sans=None,
    encoding="PEM",
):
    if csr:
        csr = x509.load_pem_x509_csr(str.encode(csr), default_backend())
    elif key:
        key = serialization.load_pem_private_key(
            str.encode(key), password=None, backend=default_backend()
        )
        csr = _create_csr(common_name, key, oids, sans)
    else:
        key = _create_key(common_name, public_exponent, key_size)
        csr = _create_csr(common_name, key, oids, sans)
    modhash = _create_modhash(csr)
    if isinstance(key, _RSAPrivateKey):
        key = key.private_bytes(
            encoding=ENCODING[encoding],
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode("utf-8")
    if isinstance(csr, _CertificateSigningRequest):
        csr = csr.public_bytes(ENCODING[encoding]).decode("utf-8")
    return (modhash, key, csr)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json


def jsonify(string):
    return json.loads(str(string))


def json_print(obj, *exclusions):
    [obj.pop(exclusion, None) for exclusion in exclusions]
    print(json.dumps(obj, indent=2, sort_keys=True))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from datetime import datetime


def utcnow():
    return datetime.utcnow()


def datetime2int(dt, pattern="%Y%m%d%H%M%S"):
    return int(dt.strftime(pattern))


def int2datetime(i, pattern="%Y%m%d%H%M%S"):
    return datetime.strptime(str(i), pattern)


def string2datetime(s, pattern="%Y-%m-%d"):
    return datetime.strptime(s, pattern)


def string2int(s, pattern="%Y-%m-%d"):
    return datetime2int(string2datetime(s, pattern=pattern))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import fnmatch


def regexes(items, includes=None, excludes=None):
    return _sift(items, includes, excludes, _include_regex, _exclude_regex)


def fnmatches(items, includes=None, excludes=None):
    if "*" in includes and not excludes:
        return items
    return _sift(items, includes, excludes, _include_fnmatch, _exclude_fnmatch)


def globs(dirpath, includes=None, excludes=None):
    items = os.listdir(dirpath)
    return fnmatches(items, includes, excludes, _include_fnmatch, _exclude_fnmatch)


def _sift(items, includes, excludes, include_fn, exclude_fn):
    items = [item for item in items if include_fn(item, includes)]
    items = [item for item in items if exclude_fn(item, excludes)]
    return items


def _include_fnmatch(item, includes):
    if not includes:
        includes = ["*"]
    return any([fnmatch.fnmatch(item, include) for include in includes])


def _exclude_fnmatch(item, excludes):
    if not excludes:
        excludes = [""]
    return all([not fnmatch.fnmatch(item, exclude) for exclude in excludes])


def _include_regex(item, includes):
    if not includes:
        includes = [".*"]
    raise NotImplementedError


def _exclude_regex(item, excludes):
    if not excludes:
        excludes = ["()"]
    raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from copy import deepcopy


class MergeError(Exception):
    def __init__(self, message):
        super(MergeError, self).__init__(message)


class DictDoesntHaveHeadError(Exception):
    def __init__(self, d):
        keys = list(d.keys())
        message = f"dictionary.keys(): {keys} does not have a single key to be considered a head"
        super(DictDoesntHaveHeadError, self).__init__(message)


class NotDictError(Exception):
    def __init__(self, o):
        t = type(o)
        message = f"{o} is {t}, not a dict as required"
        super(NotDictError, self).__init__(message)


def dict_to_attrs(obj, d):
    for k, v in d.items():
        setattr(obj, k, v)
    return obj


def update(d1, d2):
    d1.update(d2)
    return d1


def isstr(obj):
    return isinstance(obj, str)


def isint(obj):
    return isinstance(obj, int)


def isfloat(obj):
    return isinstance(obj, float)


def isscalar(obj):
    return (obj is None) or isstr(obj) or isint(obj) or isfloat(obj)


def _merge(obj1, obj2):
    """merges obj2 into obj1 and return merged result

    NOTE: tuples and arbitrary objects are not handled as it is totally
    ambiguous what should happen"""
    key = None
    result = deepcopy(obj1)
    try:
        if isscalar(result):
            result = obj2
        elif isinstance(result, list):
            if isinstance(obj2, list):
                result.extend(obj2)
            else:
                result.append(obj2)
        elif isinstance(result, dict):
            if isinstance(obj2, dict):
                for key in obj2:
                    if key in result:
                        result[key] = _merge(result[key], obj2[key])
                    else:
                        result[key] = obj2[key]
            else:
                raise MergeError(
                    'Cannot merge non-dict "%s" into dict "%s"' % (obj2, result)
                )
        else:
            raise MergeError('NOT IMPLEMENTED "%s" into "%s"' % (obj2, result))
    except TypeError as e:
        raise MergeError(
            f'TypeError "{e}" in key "{key}" when merging "{obj2}" into "{result}"'
        )
    return result


def merge(*objs):
    if len(objs) == 0:
        return {}
    elif len(objs) == 1:
        return objs[0]
    result = objs[0]
    for obj in objs[1:]:
        result = _merge(result, obj)
    return result


def head(d):
    if not isinstance(d, dict):
        raise NotDictError(d)
    keys = list(d.keys())
    if len(keys) == 1:
        return keys[0]
    raise DictDoesntHaveHeadError(d)


def body(d):
    return d[head(d)]


def head_body(d):
    head_ = head(d)
    return head_, d[head_]


def keys_ending(d, suffix):
    return [k for k in d.keys() if k.endswith(suffix)]


def dictify(items, sep=":"):
    result = {}
    if items:
        for item in items:
            key, value = item.split(sep)
            result[key] = result.get(key, []) + [value]
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import aiohttp
import asyncio

from json import dumps as json_dumps
from attrdict import AttrDict
from datetime import datetime
from urllib.parse import urlparse
from fnmatch import fnmatch

from .newline import windows2unix
from .singleton import Singleton


class RaiseIfError(Exception):
    def __init__(self, call):
        message = f"raise if error with {call}"
        super(RaiseIfError, self).__init__(message)


class ConflictingProxyEnv(Exception):
    def __init__(self, key, values):
        message = f"too many proxy values={values} for key={key} in env"
        super(ConflictingProxyEnv, self).__init__(message)


def get_proxy_value_from_env(key):
    values = [os.environ.get(key.lower(), None), os.environ.get(key.upper(), None)]
    values = list(set([value for value in values if value is not None]))
    if len(values) == 2:
        raise ConflictingProxyEnv(key, values)
    return values[0] if values else None


def ensure_http(url):
    if url:
        p = urlparse(url)
        if p.scheme in ("", "https"):
            return "http://" + url
    return url


class AsyncRequests(Singleton):
    def __init__(self):
        self.calls = []
        self._loop = asyncio.get_event_loop()
        self.http_proxy = ensure_http(get_proxy_value_from_env("http_proxy"))
        self.https_proxy = ensure_http(get_proxy_value_from_env("https_proxy"))
        self.no_proxy = get_proxy_value_from_env("no_proxy")
        self.no_proxies = re.split("[, ]+", self.no_proxy) if self.no_proxy else []

    @property
    def call(self):
        if self.calls:
            return self.calls[-1]
        return None

    def proxy(self, url):
        p = urlparse(url)
        if any([fnmatch(p.hostname, no_proxy) for no_proxy in self.no_proxies]):
            return None
        return {"http": self.http_proxy, "https": self.https_proxy}[p.scheme]

    async def _request(
        self,
        method,
        url=None,
        auth=None,
        headers=None,
        json=None,
        raise_if=None,
        raise_ex=None,
        repeat_if=None,
        repeat_wait=3,
        repeat_delta=None,
        verify_ssl=True,
        **kwargs,
    ):

        start = datetime.now()
        connector = aiohttp.TCPConnector(verify_ssl=verify_ssl)
        async with aiohttp.ClientSession(
            connector=connector, loop=self._loop
        ) as session:
            repeat = 0
            while True:
                send_datetime = datetime.utcnow()
                proxy = self.proxy(url)
                async with session.request(
                    method,
                    url,
                    headers=headers,
                    proxy=proxy,
                    auth=aiohttp.helpers.BasicAuth(*auth),
                    data=json_dumps(json) if json else None,
                    **kwargs,
                ) as response:

                    text = await response.text()
                    recv_datetime = datetime.utcnow()
                    send = AttrDict(
                        method=method,
                        url=url,
                        json=json,
                        proxy=proxy,
                        headers=headers,
                        datetime=send_datetime,
                    )
                    json = None
                    if response.headers["Content-Type"] == "application/json":
                        json = await response.json()
                    recv = AttrDict(
                        headers=dict(response.headers.items()),
                        status=response.status,
                        text=windows2unix(text),
                        json=json,
                        repeat=repeat,
                        datetime=datetime.utcnow(),
                    )
                    call = AttrDict(response=response, send=send, recv=recv)
                self.calls += [call]
                if repeat_if and repeat_if(call):
                    delta = datetime.now() - start
                    if repeat_delta and delta < repeat_delta:
                        repeat += 1
                        print(f"{delta} < {repeat_delta}; repeat {repeat}")
                        await asyncio.sleep(repeat_wait)
                        continue
                if raise_if and raise_if(call):
                    if raise_ex:
                        raise raise_ex(call)
                    raise RaiseIfError(call)
                break
        return call

    def request(self, method, **kw):
        return self._loop.run_until_complete(self._request(method, **kw))

    def get(self, **kw):
        return self.request("GET", **kw)

    def put(self, **kw):
        return self.request("PUT", **kw)

    def post(self, **kw):
        return self.request("POST", **kw)

    def delete(self, **kw):
        return self.request("DELETE", **kw)

    async def _requests(self, method, *kws):
        futures = [asyncio.ensure_future(self._request(method, **kw)) for kw in kws]
        responses = await asyncio.gather(*futures)
        return responses

    def requests(self, method, *kws):
        future = asyncio.ensure_future(self._requests(method, *kws))
        return self._loop.run_until_complete(future)

    def gets(self, *kws):
        return self.requests("GET", *kws)

    def puts(self, *kws):
        return self.requests("PUT", *kws)

    def posts(self, *kws):
        return self.requests("POST", *kws)

    def deletes(self, *kws):
        return self.requests("DELETE", *kws)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import inspect


def docstr(*args, **kwargs):
    frame = inspect.currentframe().f_back
    funcname = frame.f_code.co_name
    globals_ = frame.f_globals
    func = globals_[funcname]
    func.__doc__ = func.__doc__.format(*args, **kwargs)
    return func.__doc__



# -*- coding: utf-8 -*-
"""
config
"""

import os
import re
import pwd
import sys
import time
import logging
import sh

from decouple import UndefinedValueError, AutoConfig, config

LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]

LOG_LEVEL = config("LOG_LEVEL", logging.WARNING, cast=int)

logging.basicConfig(
    stream=sys.stdout, level=LOG_LEVEL, format="%(asctime)s %(name)s %(message)s"
)
logging.Formatter.converter = time.gmtime
log = logging.getLogger(__name__)


class NotGitRepoError(Exception):
    """
    NotGitRepoError
    """

    def __init__(self):
        """
        init
        """
        msg = "not a git repository error"
        super(NotGitRepoError, self).__init__(msg)


class NoGitRepoOrEnvError(Exception):
    """
    NoGitRepoOrEnvError
    """

    def __init__(self):
        """
        init
        """
        msg = "no git repo or env found error"
        super(NoGitRepoOrEnvError, self).__init__(msg)


def git(*args, strip=True, **kwargs):
    """
    git
    """
    try:
        result = str(sh.contrib.git(*args, **kwargs))  # pylint: disable=no-member
        if strip:
            result = result.strip()
        return result
    except sh.ErrorReturnCode as e:
        stderr = e.stderr.decode("utf-8")
        if "not a git repository" in stderr.lower():
            raise NotGitRepoError
        log.error(e)


class AutoConfigPlus(AutoConfig):  # pylint: disable=too-many-public-methods
    """
    thin wrapper around AutoConfig adding some extra features
    """

    @property
    def APP_UID(self):
        """
        uid
        """
        return os.getuid()

    @property
    def APP_GID(self):
        """
        gid
        """
        return pwd.getpwuid(self.APP_UID).pw_gid

    @property
    def APP_USER(self):
        """
        user
        """
        return pwd.getpwuid(self.APP_UID).pw_name

    @property
    def APP_PORT(self):
        """
        port
        """
        return self("APP_PORT", 5000, cast=int)

    @property
    def APP_JOBS(self):
        """
        jobs
        """
        try:
            return call("nproc")[1].strip()
        except:  # pylint: disable=bare-except
            return 1

    @property
    def APP_TIMEOUT(self):
        """
        timeout
        """
        return self("APP_TIMEOUT", 120, cast=int)

    @property
    def APP_WORKERS(self):
        """
        workers
        """
        return self("APP_WORKERS", 2, cast=int)

    @property
    def APP_MODULE(self):
        """
        module
        """
        return self("APP_MODULE", "main:app")

    @property
    def APP_REPOROOT(self):
        """
        reporoot
        """
        return git("rev-parse", "--show-toplevel")

    @property
    def APP_INSTALLPATH(self):
        """
        install path
        """
        return self("APP_INSTALLPATH", "/usr/src/app")

    @property
    def APP_VERSION(self):
        """
        version
        """
        try:
            return git("describe", "--abbrev=7", "--always")
        except NotGitRepoError:
            return self("APP_VERSION")

    @property
    def APP_BRANCH(self):
        """
        branch
        """
        try:
            return git("rev-parse", "--abbrev-ref", "HEAD")
        except NotGitRepoError:
            return self("APP_BRANCH")

    @property
    def APP_DEPENV(self):
        """
        deployment environment
        """
        branch = self.APP_BRANCH
        if branch == "master":
            return "prod"
        elif branch.startswith("stage/"):
            return "stage"
        return "dev"

    @property
    def APP_SRCTAR(self):
        """
        srctar
        """
        try:
            return self("APP_SRCTAR")
        except UndefinedValueError:
            return ".src.tar.gz"

    @property
    def APP_REVISION(self):
        """
        revision
        """
        try:
            return git("rev-parse", "HEAD")
        except NotGitRepoError:
            return self("APP_REVISION")

    @property
    def APP_REMOTE_ORIGIN_URL(self):
        """
        remote origin url
        """
        try:
            return git("config", "--get", "remote.origin.url")
        except NotGitRepoError:
            return self("APP_REMOTE_ORIGIN_URL")

    @property
    def APP_REPONAME(self):
        """
        reponame
        """
        pattern = r"((ssh|https)://)?(git@)?github.com[:/](?P<reponame>[A-Za-z0-9\/\-_]+)(.git)?"
        match = re.search(pattern, self.APP_REMOTE_ORIGIN_URL)
        return match.group("reponame")

    @property
    def APP_PROJNAME(self):
        """
        projname
        """
        return os.path.basename(self.APP_REPONAME)

    @property
    def APP_PROJPATH(self):
        """
        projpath
        """
        return os.path.join(self.APP_REPOROOT, self.APP_PROJNAME)

    @property
    def APP_BOTPATH(self):
        """
        botpath
        """
        return os.path.join(self.APP_PROJPATH, "bot")

    @property
    def APP_DBPATH(self):
        """
        dbpath
        """
        return os.path.join(self.APP_PROJPATH, "db")

    @property
    def APP_TESTPATH(self):
        """
        testpath
        """
        return os.path.join(self.APP_PROJPATH, "tests")

    @property
    def APP_LS_REMOTE(self):
        """
        ls-remote
        """
        try:
            result = git("ls-remote", f"https://github.com/{self.APP_REPONAME}")
        except NotGitRepoError:
            result = self("APP_LS_REMOTE")
        return {
            refname: revision
            for revision, refname in [line.split() for line in result.split("\n")]
        }

    @property
    def APP_GSM_STATUS(self):
        """
        gsm status
        """
        try:
            result = git("submodule", "status", strip=False)
        except NotGitRepoError:
            result = self("APP_GSM_STATUS")
        pattern = r"([ +-])([a-f0-9]{40}) ([A-Za-z0-9\/\-_.]+)( .*)?"
        matches = re.findall(pattern, result)
        states = {
            " ": True,  # submodule is checked out the correct revision
            "+": False,  # submodule is checked out to a different revision
            "-": None,  # submodule is not checked out
        }
        return {
            repopath: [revision, states[state]]
            for state, revision, repopath, _ in matches
        }

    def __getattr__(self, attr):
        """
        getattr
        """
        log.info(f"attr = {attr}")
        if attr == "create_doit_tasks":  # note: to keep pydoit's hands off
            return lambda: None
        result = self(attr)
        try:
            return int(result)
        except ValueError:
            return result


CFG = AutoConfigPlus()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys

from .json import json_print
from .yaml import yaml_print

OUTPUT = ["json", "yaml"]


def output_print(json, output):
    if output == "yaml":
        yaml_print(json)
    elif output == "json":
        json_print(json)


def default_output():
    return OUTPUT[int(sys.stdout.isatty())]
